'// DupeFind - a duplicate file location and removal tool
'// Copyright 2004-2008, Casey Sprangel, all rights reserved.
'// Written in PowerBasic/CC 4.0 (http://www.powerbasic.com)
'//---------------------------------------------------------------------------
'// Possible additions:
'//  + External control over erase preferences (keep older/newer dates, shorter/longer file names, deep/shallow paths, etc.)
'//  + Command line control over directory recursion
'//  + Optional user prompting and choice for each file delete
'//---------------------------------------------------------------------------
%myver = 20080324
#INCLUDE "WIN32API.INC"                                       '// include our API calls for fast disk-io
'//---------------------------------------------------------------------------

DECLARE FUNCTION fkill(a AS STRING) AS LONG                   '// does the work of erasing files
DECLARE FUNCTION exist(filen AS STRING) AS LONG               '// checks to see if a file exists or not
DECLARE FUNCTION same(f1 AS STRING, f2 AS STRING) AS LONG     '// checks to see if two files are exactly the same inside
DECLARE FUNCTION pshow() AS LONG                              '// updates the status bar
DECLARE FUNCTION dpush (a AS STRING) AS LONG                  '// push a directory name on to the stack
DECLARE FUNCTION fpush (a AS STRING) AS LONG                  '// push a file name on to the stack
DECLARE FUNCTION FileDateTime(ft AS FileTime) AS STRING       '// date and time translating
DECLARE FUNCTION flog(a AS STRING) AS LONG                    '// writes to the log file
DECLARE FUNCTION Crc32(BYVAL Address AS DWORD, BYVAL Length AS LONG, BYVAL Seed AS LONG) AS LONG '// calculate 32 bit crc on a chunk of data
DECLARE FUNCTION crcfile(f1 AS STRING) AS STRING              '// gets 32 bit crc of a file

GLOBAL dp AS LONG                                             '// pointer to the total number of directories we know about
GLOBAL d() AS STRING                                          '// holds all of the path names we know about
GLOBAL fp AS LONG                                             '// pointer to the total number of file names we have in our table
GLOBAL f() AS STRING                                          '// holds all of the files we find on the drive
GLOBAL logfile AS STRING                                      '// the name of our log file
GLOBAL dupekill AS LONG                                       '// flag to show if we actually erased files or not
GLOBAL verbose AS LONG                                        '// flag to show extra processing information on the screen
GLOBAL mqueue AS STRING                                       '// holds log file headers to see if we need to log anything

$longform = "0,"                                              '// generic comma delim number format
$dflag = CHR$(1)                                              '// the delete flag character (faster to overwrite a byte than erase a whole string)
'//                                1    2    3                    4    5
'// Unformats file string. format: size|name|last_write_date_time|path|crc
MACRO unf(a) = d( VAL(PARSE$(a,"|",4) ) ) + "\" + PARSE$(a, "|", 2) + " " + PARSE$(a, "|", 1) + " " + PARSE$(a, "|", 3)

MACRO flen(a) = LEN(a) - INSTR(-1, a, ANY ":\")               '// find last slash mark or colon character

MACRO fsize(h, L) = CDBL(h) * CDBL(&H10000) * CDBL(&H10000) + CDBL(L) '// calculate the file size

MACRO tense = IIF$(dupekill, " was ", " would have been ")    '// used to show user messages based on if we delete files or not

MACRO debug(verbose_text)                                     '// show debug output only if in verbose mode
 IF verbose THEN
   STDOUT "[V] " + verbose_text                               '// show on local screen
   flog "[V] " + verbose_text + $CRLF                         '// include in log file
 END IF
END MACRO

'//---------------------------------------------------------------------------

FUNCTION Crc32(BYVAL Address AS DWORD, BYVAL Length AS LONG, BYVAL Seed AS LONG) AS LONG
  ! push    EBX                                               ;// save EBX for PowerBASIC
  ! push    EDI                                               ;// save EDI for PowerBASIC
  ! mov     EDI,Address                                       ;// address in EDI
  ! mov     ECX,Length                                        ;// length in ECX
  ! jecxz   CrcDone                                           ;// exit is zero length
  ! cld                                                       ;// clear the direction flag

BuildCRC:
  ! movzx   EBX, Byte Ptr [EDI]                               ;// get a char
  ! mov     AX, Seed[1]                                       ;// get 2nd and 3rd bytes of seed
  ! xor     DX, DX                                            ;// clear DX
  ! mov     DL, Seed[3]                                       ;// get 4th byte of seed
  ! xor     BL, Seed[0]                                       ;// xor char against first byte of seed
  ! xor     BH, BH                                            ;// clear BH
  ! shl     BX, 2                                             ;// shift the index
  ! xor     AX, CrcTable[EBX]                                 ;// xor low-half against the table
  ! xor     DX, CrcTable[EBX+2]                               ;// xor high-half against the table
  ! mov     Seed[0], AX                                       ;// save the result
  ! mov     Seed[2], DX                                       ;//   ...
  ! inc     EDI                                               ;// move to next char
  ! loop    BuildCRC                                          ;// do ECX times

CrcDone:
  ! pop     EDI                                               ;// restore EDI
  ! pop     EBX                                               ;// restore EBX
  FUNCTION = Seed                                             '// return the crc value
  EXIT FUNCTION                                               '// leave, we are done

CrcTable:
  ! dd &H000000000&, &H077073096&, &H0EE0E612C&, &H0990951BA&
  ! dd &H0076DC419&, &H0706AF48F&, &H0E963A535&, &H09E6495A3&
  ! dd &H00EDB8832&, &H079DCB8A4&, &H0E0D5E91E&, &H097D2D988&
  ! dd &H009B64C2B&, &H07EB17CBD&, &H0E7B82D07&, &H090BF1D91&
  ! dd &H01DB71064&, &H06AB020F2&, &H0F3B97148&, &H084BE41DE&
  ! dd &H01ADAD47D&, &H06DDDE4EB&, &H0F4D4B551&, &H083D385C7&
  ! dd &H0136C9856&, &H0646BA8C0&, &H0FD62F97A&, &H08A65C9EC&
  ! dd &H014015C4F&, &H063066CD9&, &H0FA0F3D63&, &H08D080DF5&
  ! dd &H03B6E20C8&, &H04C69105E&, &H0D56041E4&, &H0A2677172&
  ! dd &H03C03E4D1&, &H04B04D447&, &H0D20D85FD&, &H0A50AB56B&
  ! dd &H035B5A8FA&, &H042B2986C&, &H0DBBBC9D6&, &H0ACBCF940&
  ! dd &H032D86CE3&, &H045DF5C75&, &H0DCD60DCF&, &H0ABD13D59&
  ! dd &H026D930AC&, &H051DE003A&, &H0C8D75180&, &H0BFD06116&
  ! dd &H021B4F4B5&, &H056B3C423&, &H0CFBA9599&, &H0B8BDA50F&
  ! dd &H02802B89E&, &H05F058808&, &H0C60CD9B2&, &H0B10BE924&
  ! dd &H02F6F7C87&, &H058684C11&, &H0C1611DAB&, &H0B6662D3D&
  ! dd &H076DC4190&, &H001DB7106&, &H098D220BC&, &H0EFD5102A&
  ! dd &H071B18589&, &H006B6B51F&, &H09FBFE4A5&, &H0E8B8D433&
  ! dd &H07807C9A2&, &H00F00F934&, &H09609A88E&, &H0E10E9818&
  ! dd &H07F6A0DBB&, &H0086D3D2D&, &H091646C97&, &H0E6635C01&
  ! dd &H06B6B51F4&, &H01C6C6162&, &H0856530D8&, &H0F262004E&
  ! dd &H06C0695ED&, &H01B01A57B&, &H08208F4C1&, &H0F50FC457&
  ! dd &H065B0D9C6&, &H012B7E950&, &H08BBEB8EA&, &H0FCB9887C&
  ! dd &H062DD1DDF&, &H015DA2D49&, &H08CD37CF3&, &H0FBD44C65&
  ! dd &H04DB26158&, &H03AB551CE&, &H0A3BC0074&, &H0D4BB30E2&
  ! dd &H04ADFA541&, &H03DD895D7&, &H0A4D1C46D&, &H0D3D6F4FB&
  ! dd &H04369E96A&, &H0346ED9FC&, &H0AD678846&, &H0DA60B8D0&
  ! dd &H044042D73&, &H033031DE5&, &H0AA0A4C5F&, &H0DD0D7CC9&
  ! dd &H05005713C&, &H0270241AA&, &H0BE0B1010&, &H0C90C2086&
  ! dd &H05768B525&, &H0206F85B3&, &H0B966D409&, &H0CE61E49F&
  ! dd &H05EDEF90E&, &H029D9C998&, &H0B0D09822&, &H0C7D7A8B4&
  ! dd &H059B33D17&, &H02EB40D81&, &H0B7BD5C3B&, &H0C0BA6CAD&
  ! dd &H0EDB88320&, &H09ABFB3B6&, &H003B6E20C&, &H074B1D29A&
  ! dd &H0EAD54739&, &H09DD277AF&, &H004DB2615&, &H073DC1683&
  ! dd &H0E3630B12&, &H094643B84&, &H00D6D6A3E&, &H07A6A5AA8&
  ! dd &H0E40ECF0B&, &H09309FF9D&, &H00A00AE27&, &H07D079EB1&
  ! dd &H0F00F9344&, &H08708A3D2&, &H01E01F268&, &H06906C2FE&
  ! dd &H0F762575D&, &H0806567CB&, &H0196C3671&, &H06E6B06E7&
  ! dd &H0FED41B76&, &H089D32BE0&, &H010DA7A5A&, &H067DD4ACC&
  ! dd &H0F9B9DF6F&, &H08EBEEFF9&, &H017B7BE43&, &H060B08ED5&
  ! dd &H0D6D6A3E8&, &H0A1D1937E&, &H038D8C2C4&, &H04FDFF252&
  ! dd &H0D1BB67F1&, &H0A6BC5767&, &H03FB506DD&, &H048B2364B&
  ! dd &H0D80D2BDA&, &H0AF0A1B4C&, &H036034AF6&, &H041047A60&
  ! dd &H0DF60EFC3&, &H0A867DF55&, &H0316E8EEF&, &H04669BE79&
  ! dd &H0CB61B38C&, &H0BC66831A&, &H0256FD2A0&, &H05268E236&
  ! dd &H0CC0C7795&, &H0BB0B4703&, &H0220216B9&, &H05505262F&
  ! dd &H0C5BA3BBE&, &H0B2BD0B28&, &H02BB45A92&, &H05CB36A04&
  ! dd &H0C2D7FFA7&, &H0B5D0CF31&, &H02CD99E8B&, &H05BDEAE1D&
  ! dd &H09B64C2B0&, &H0EC63F226&, &H0756AA39C&, &H0026D930A&
  ! dd &H09C0906A9&, &H0EB0E363F&, &H072076785&, &H005005713&
  ! dd &H095BF4A82&, &H0E2B87A14&, &H07BB12BAE&, &H00CB61B38&
  ! dd &H092D28E9B&, &H0E5D5BE0D&, &H07CDCEFB7&, &H00BDBDF21&
  ! dd &H086D3D2D4&, &H0F1D4E242&, &H068DDB3F8&, &H01FDA836E&
  ! dd &H081BE16CD&, &H0F6B9265B&, &H06FB077E1&, &H018B74777&
  ! dd &H088085AE6&, &H0FF0F6A70&, &H066063BCA&, &H011010B5C&
  ! dd &H08F659EFF&, &H0F862AE69&, &H0616BFFD3&, &H0166CCF45&
  ! dd &H0A00AE278&, &H0D70DD2EE&, &H04E048354&, &H03903B3C2&
  ! dd &H0A7672661&, &H0D06016F7&, &H04969474D&, &H03E6E77DB&
  ! dd &H0AED16A4A&, &H0D9D65ADC&, &H040DF0B66&, &H037D83BF0&
  ! dd &H0A9BCAE53&, &H0DEBB9EC5&, &H047B2CF7F&, &H030B5FFE9&
  ! dd &H0BDBDF21C&, &H0CABAC28A&, &H053B39330&, &H024B4A3A6&
  ! dd &H0BAD03605&, &H0CDD70693&, &H054DE5729&, &H023D967BF&
  ! dd &H0B3667A2E&, &H0C4614AB8&, &H05D681B02&, &H02A6F2B94&
  ! dd &H0B40BBE37&, &H0C30C8EA1&, &H05A05DF1B&, &H02D02EF8D&
END FUNCTION

'//---------------------------------------------------------------------------

FUNCTION show(d AS LONG, r AS LONG) AS LONG                   '// put stuff on the console title bar
  STATIC d2 AS LONG                                           '// holds last number of duplicate files found
  STATIC t AS STRING                                          '// holds the timer second value
  IF t <> RIGHT$(TIME$,1) THEN                                '// do we need to update the title bar yet?
    d2 = -1                                                   '// flag to show that we need to update
  END IF
  IF d2 <> d THEN                                             '// do we need to update the title bar?
    CONSOLE NAME "Found " + FORMAT$(d, $longform) + " duplicate" + IIF$(d = 1,"","s") + ". " + FORMAT$(r, $longform) + " to be checked."
    d2 = d                                                    '// remember the number of duplicate files shown
    t = RIGHT$(TIME$,1)                                       '// remember this second to detect changes later
  END IF
END FUNCTION

'//---------------------------------------------------------------------------

FUNCTION fkill(a AS STRING) AS LONG                           '// this does the work of actually erasing files we don't want
  '// Returns -1 if file is erased (real or simulated)
  IF NOT exist(a) THEN                                        '// file does not exist?
    debug ("File erase: " + $DQ + a + $DQ + " does not exist.")
    FUNCTION = -1                                             '// return true (success)
    EXIT FUNCTION                                             '// leave, we are done
  END IF
  IF dupekill THEN                                            '// can we erase a file?
    debug ("File erase: " + $DQ + a + $DQ)
    ON ERROR RESUME NEXT                                      '// turn local error traps off
    KILL a                                                    '// delete the file
    ON ERROR GOTO 0                                           '// restore normal error trap handling
    FUNCTION = NOT exist(a)                                   '// return true if file does not exist
  END IF
END FUNCTION

'//---------------------------------------------------------------------------

FUNCTION exist(filen AS STRING) AS LONG                       '// check if a file exists
  '// Returns -1 if a file or path exists.
  '// Returns 0 if it does not exist.
  '// Does not change DIR$ FindFirst/FindNext pointers.
  LOCAL result AS LONG                                        '// a junk variable, we don't care what comes back
  ON ERROR RESUME NEXT                                        '// turn off error traps for this function
  IF RIGHT$(filen, 1) = ":" THEN                              '// does file name end in a colon (i.e. a drive letter)?
    result = GETATTR(filen + "\")                             '// see if the root path exists (adding a slash)
  ELSE                                                        '// this is a normal path or filename
    result = GETATTR(RTRIM$(filen, "\"))                      '// see if it exists, paths with trailing slash marks are trimmed
  END IF
  ON ERROR GOTO 0
  FUNCTION = (ERRCLEAR = 0)                                   '// return true if there was no error
END FUNCTION

'//---------------------------------------------------------------------------

FUNCTION same(f1 AS STRING, f2 AS STRING) AS LONG             '// A simple file compare to see if the files are duplicates
  '// Returns -1 if files are exactly the same
  LOCAL h1 AS LONG                                            '// file handle one
  LOCAL h2 AS LONG                                            '// file handle two
  LOCAL buf1 AS STRING                                        '// file one buffer
  LOCAL buf2 AS STRING                                        '// file two buffer
  LOCAL bsize AS LONG                                         '// maximum size of input buffer
  LOCAL i AS QUAD                                             '// this tracks to the file offset - can get very large!
  LOCAL result AS LONG

  IF NOT exist(f1) THEN                                       '// file not here?
    debug ("SAME: " + $DQ + f1 + $DQ + " not found.")
    EXIT FUNCTION                                             '// leave, it can't be a dupe
  END IF
  IF NOT exist(f2) THEN                                       '// file not here?
    debug ("SAME: " + $DQ + f2 + $DQ + " not found.")
    EXIT FUNCTION                                             '// leave, it can't be a dupe
  END IF
  h1 = FREEFILE                                               '// get next available file handle
  OPEN f1 FOR BINARY ACCESS READ AS h1                        '// open the file in read only mode
  IF LOF(#h1) = 0 THEN                                         '// empty file?
    debug ("SAME: File size 1 (" + f1 + ") is " + FORMAT$(LOF(h1)))
    CLOSE h1
    EXIT FUNCTION
  END IF
  h2 = FREEFILE                                               '// get next available file handle
  OPEN f2 FOR BINARY ACCESS READ AS h2                        '// open the file in read only mode
  debug ("SAME: File size 1 (" + f1 + ") is " + FORMAT$(LOF(h1)))
  debug ("SAME: File size 2 (" + f2 + ") is " + FORMAT$(LOF(h2)))
  IF LOF(#h1) <> LOF(#h2) THEN                                '// are file lengths different or empty? we can take a shortcut exit
    debug ("SAME: File sizes do not match.")
    CLOSE h1                                                  '// close the file and release the handle
    CLOSE h2                                                  '// close the file and release the handle
    EXIT FUNCTION                                             '// leave, they are not the same
  END IF
  bsize = 1048576                                             '// set the default buffer size in bytes
'  bsize = 4096
  IF bsize > LOF(h1) THEN                                     '// is buffer size more than the size of the file?
    bsize = LOF(h1)                                           '// set buffer size to read the whole file in one chunk
  END IF
  i = 0                                                       '// start our counter at zero
  DO WHILE i < LOF(h1)                                        '// loop through the whole file
    GET$ #h1, bsize, buf1                                     '// get a chunk of the first file
    GET$ #h2, bsize, buf2                                     '// get a chunk of the second file
    IF buf1 <> buf2 THEN                                      '// is the file not the same so far?
      INCR result                                             '// set the flag
      EXIT LOOP                                               '// exit the loop, we are done reading
    END IF
    i = i + bsize                                             '// add length of the buffer to our counter
  LOOP
  CLOSE h1                                                    '// close the file and release the handle
  CLOSE h2                                                    '// close the file and release the handle
  IF result = 0 THEN                                          '// no mismatches found?
    debug ("SAME: Files match.")
    FUNCTION = -1                                             '// true, files are the same
  END IF
END FUNCTION

'//---------------------------------------------------------------------------

FUNCTION crcfile(f1 AS STRING) AS STRING                      '// reads a file and returns a CRC
  LOCAL h1 AS LONG                                            '// file handle one
  LOCAL buf1 AS STRING                                        '// file one buffer
  LOCAL bsize AS LONG                                         '// maximum size of input buffer
  LOCAL i AS QUAD                                             '// this tracks to the file offset - can get very large!
  LOCAL crc AS LONG

  IF NOT exist(f1) THEN                                       '// file not here?
    EXIT FUNCTION                                             '// leave, it can't be a dupe
  END IF
  h1 = FREEFILE                                               '// get next available file handle
  OPEN f1 FOR BINARY ACCESS READ AS h1                        '// open the file in read only mode
  bsize = 4096                                                '// set the default buffer size in bytes
  IF bsize > LOF(h1) THEN                                     '// is buffer size more than the size of the file?
    bsize = LOF(h1)                                           '// set buffer size to read the whole file in one chunk
  END IF
  i = 0                                                       '// start our counter at zero
  DO WHILE i < LOF(h1)                                        '// loop through the whole file
    GET$ #h1, bsize, buf1                                     '// get a chunk of the first file
    Crc = crc + Crc32(STRPTR(buf1), LEN(buf1), -1)            '// get the crc for this chunk of the file
    i = i + bsize                                             '// add length of the buffer to our counter
  LOOP
  CLOSE h1                                                    '// close the file and release the handle
  Crc = Crc XOR -1&                                           '// compute PKZIP/ZMODEM compatible inverse CRC result
  FUNCTION = FORMAT$(crc)
END FUNCTION

'//---------------------------------------------------------------------------

FUNCTION pshow() AS LONG                                      '// show progress on console title bar
  STATIC pshow_t AS STRING                                    '// remember this value between function calls
  IF pshow_t <> RIGHT$(TIME$, 1) THEN                         '// did the time change?
    CONSOLE NAME "Found " + FORMAT$(fp, $longform) + " files : " + FORMAT$(dp, $longform) + " directories. " + d(dp) '// change the title bar to show progress
    pshow_t = RIGHT$(TIME$, 1)                                '// remember the time
  END IF
END FUNCTION

'//---------------------------------------------------------------------------

FUNCTION dpush (a AS STRING) AS LONG                          '// push a directory name on to the stack
  IF LEN(a) = 0 THEN                                          '// is this parameter empty?
    EXIT FUNCTION                                             '// leave this function, nothing to add
  END IF
  IF a = "." OR a = ".." THEN                                 '// no placeholders allowed
    EXIT FUNCTION                                             '// leave the function, can't add these
  END IF
  INCR dp                                                     '// add one to our stack pointer
  IF dp > UBOUND(d()) THEN                                    '// is pointer past the end of our table?
    REDIM PRESERVE d(1000 + dp)                               '// add more elements to our table
  END IF
  d(dp) = RTRIM$(a,"\")                                       '// no trailing slash marks on the directory name
  pshow                                                       '// update the status bar
END FUNCTION

'//---------------------------------------------------------------------------

FUNCTION fpush (a AS STRING) AS LONG                          '// push a file name on to the stack
  IF LEN(a) = 0 THEN                                          '// is this parameter empty?
    EXIT FUNCTION                                             '// leave this function, nothing to add
  END IF
  INCR fp                                                     '// add one to our stack pointer
  IF fp > UBOUND(f()) THEN                                    '// are we pointing past the end of our table?
    REDIM PRESERVE f(1000 + fp)                               '// add more elements to our table
  END IF
  f(fp) = a                                                   '// put this element on to our table
  pshow                                                       '// update the status bar
END FUNCTION

'//---------------------------------------------------------------------------

FUNCTION FileDateTime(ft AS FileTime) AS STRING
  LOCAL st AS SystemTime                                      '// structure for working with Windows time formats
  LOCAL zText AS ASCIIZ * 255                                 '// buffer text to hold our answers
  LOCAL Temp AS STRING                                        '// holds our answer part way through
  FileTimeToLocalFileTime ft, ft                              '// Convert the file time from UTC to local time
  FileTimeToSystemTime ft, st                                 '// Convert the file time into a compatible system time
  GetDateFormat %LOCALE_USER_DEFAULT, %DATE_SHORTDATE, st, BYVAL %NULL, zText, 255 '// Create a date string using the local settings
  Temp = zText
  GetTimeFormat %LOCALE_USER_DEFAULT, %TIME_NOSECONDS, st, BYVAL %NULL, zText, 255 '// Create a time string using the local settings
  FUNCTION = Temp + " " + zText                               '// Return the file date and time
END FUNCTION

'//---------------------------------------------------------------------------

FUNCTION flog(a AS STRING) AS LONG                            '// file log
  LOCAL i AS LONG                                             '// holds temp file handle
  i = FREEFILE                                                '// get next available file handle
  IF LEN(mqueue) > 0 THEN                                     '// is the header still waiting to be written to the file?
    OPEN logfile FOR OUTPUT AS i                              '// open our log file for output
    ?#i, mqueue;                                              '// dump our entry to the log file
    CLOSE i                                                   '// close the file
    mqueue = ""                                               '// erase it, so we know not to do this next time
  END IF
  OPEN logfile FOR APPEND AS i                                '// open our log file for output
  ?#i, a;                                                     '// dump our entry to the log file
  CLOSE i                                                     '// close the file
END FUNCTION

'//---------------------------------------------------------------------------

FUNCTION PBMAIN()
  REGISTER i AS LONG                                          '// temp pointer (stored inside the CPU register for max speed)
  REGISTER j AS LONG                                          '// temp pointer 2
  REGISTER k AS LONG                                          '// temp pointer 3
  DIM p AS LONG                                               '// offset pointer
  DIM d(1000) AS STRING                                       '// make room for the first 1000 directories
  DIM f(1000) AS STRING                                       '// make room for the first 1000 records
  DIM g(10) AS STRING                                         '// start with room for up to 10 duplicates to compare
  DIM hFindFile AS LONG                                       '// handle to keep track of directory scanning
  DIM lpFindFileData AS WIN32_FIND_DATA                       '// structure for getting file info
  DIM buf AS STRING                                           '// temp buffer
  DIM buf2 AS STRING                                          '// temp buffer 2
  DIM dupe AS LONG                                            '// total number of duplicate files we find
  DIM title AS STRING
  DIM fqueue AS STRING                                        '// holds the main file name being compared
  DIM fpump AS STRING                                         '// holds a multi-line log entry until we do the work
  DIM freespace AS QUAD                                       '// tracks size of all files erased (i.e. free space recovered)
  DIM tstart AS LONG                                          '// start time used for measuring run time

  logfile = "DupeFind.lst"                                    '// log file name
  title = "DupeFind version " + FORMAT$(%myver)               '// program title
  STDOUT title                                                '// show the title
  buf = TRIM$(COMMAND$)                                       '// get the command line from the user
  IF INSTR(COMMAND$,"?") > 0 THEN                             '// did user ask for help?
    STDOUT "Copyright 2004-2008, Casey Sprangel, all rights reserved."
    STDOUT ""
    STDOUT "This tool helps locate duplicate files on your hard drive. This is a true"
    STDOUT "binary file comparison able to detect duplicate file even if they have"
    STDOUT "different names."
    STDOUT ""
    STDOUT "Syntax:  DUPEFIND [<path>] [/D] [/V]"
    STDOUT ""
    STDOUT "The path is the starting directory where you want to start looking for"
    STDOUT "duplicate files. All subdirectories below the starting path are scanned"
    STDOUT "also. Remember that this is a complete binary file scan. It can take a long"
    STDOUT "time if the hard drive is very full or slow."
    STDOUT ""
    STDOUT "If the /D switch is used, this allows automatic deleting of the duplicate"
    STDOUT "files as they are found. The preference is for files with longer (and"
    STDOUT "hopefully more descriptive) file names in paths that are closer to the root."
    STDOUT ""
    STDOUT "The /V switch enables verbose mode which shows extra detail about internal"
    STDOUT "processing and function."
    STDOUT ""
    STDOUT "A report file named " + logfile + " is created listing all of the duplicate"
    STDOUT "files found including the complete path to each file."
    EXIT FUNCTION
  END IF
  i = INSTR(UCASE$(buf), "/V")                                '// look for command line switch
  IF i > 0 THEN                                               '// is the switch found?
    verbose = -1                                              '// true
    STDOUT "[V] Verbose mode enabled."
    buf = TRIM$(LEFT$(buf, i-1) + MID$(buf, i + 2))           '// cut out command line switch
    tstart = TIMER                                            '// remember the start time
  END IF
  i = INSTR(UCASE$(buf), "/D")                                '// look for command line switch
  IF i > 0 THEN                                               '// is the switch found?
    dupekill = -1                                             '// true
    buf = TRIM$(LEFT$(buf, i - 1) + MID$(buf, i + 2))         '// cut out command line switch
  END IF
  IF buf = "." OR buf = "" THEN                               '// pointer to this directory or empty?
    buf = CURDIR$                                             '// get the full path to this directory
  END IF
  IF NOT exist(buf) THEN                                      '// our starting directory does not exist?
    STDOUT "Path not found: " + buf                           '// tell user we have a problem
    EXIT FUNCTION                                             '// leave program
  END IF
  STDOUT "Scanning " + buf + " ..."
  dpush buf                                                   '// push our starting directory on to the stack
  mqueue = RIGHT$(STRING$(70,"-") + "[" + DATE$ + " " + TIME$ + "]---", 70) + $CRLF + _
           title + $CRLF + "Scanning directory: " + $DQ + d(1) + $DQ + $CRLF + STRING$(70,"-") + $CRLF '// start the log file message queue
  IF verbose THEN                                             '// do we need to add verbose information to the log file header?
    mqueue = mqueue + "[V] Command line: " + $DQ + TRIM$(COMMAND$) + $DQ + $CRLF + _
             " + Verbose mode is enabled." + $CRLF + _
             " + Delete mode is " + IIF$(dupekill,"enabled","simulated") + "." + $CRLF + $CRLF
  END IF
  i = 1                                                       '// point to the first directory so we can start scanning
  DO WHILE i <= dp                                            '// work through the directory stack until we pass the directory pointer
    hFindFile = FindFirstFile( d(i) + "\*.*", lpFindFileData) '// find first file in this directory
    IF hFindFile <> %INVALID_HANDLE_VALUE THEN                '// did we find a directory entry?
      DO                                                      '// start working out loop
        IF (lpFindFileData.dwFileAttributes AND %FILE_ATTRIBUTE_DIRECTORY) = %FILE_ATTRIBUTE_DIRECTORY THEN '// is this a directory?
          IF lpFindFileData.cFileName = "." OR lpFindFileData.cFileName = ".." THEN '// did we find the placeholder entries?
            IF FindNextFile(hFindFile, lpFindFileData) = 0 THEN '// did we fail to find another file in this directory?
              EXIT LOOP                                       '// leave, no more files to find in this directory
            END IF
            ITERATE                                           '// go work on next file
          END IF
          dpush d(i) + "\" + lpFindFileData.cFileName         '// push our new directory name on to the stack
        ELSE                                                  '// this is not a directory
          IF fsize(lpFindFileData.nFileSizeHigh, lpFindFileData.nFileSizeLow) > 0 THEN '// does this file have anything in it? (ignore blank files)
            buf = FORMAT$(fsize(lpFindFileData.nFileSizeHigh, lpFindFileData.nFileSizeLow)) + "|" + _ '// file size
                  lpFindFileData.cFileName + "|" + _          '// file name
                  FileDateTime$(lpFindFileData.ftLastWriteTime) + "|" + _  '// file date time
                  FORMAT$(i)                                  '// current directory pointer
            fpush buf                                         '// push the file name on to stack
          END IF                                              '// does this file have anything in it
        END IF
        IF FindNextFile(hFindFile, lpFindFileData) = 0 THEN   '// no more files in this directory?
          EXIT LOOP                                           '// leave the loop
        END IF
      LOOP
      FindClose hFindFile                                     '// we are done with this directory, close the handle
    END IF                                                    '// did we find a directory entry
    INCR i                                                    '// move pointer to next directory
    IF LEN(d(i)) = 0 THEN                                     '// did we reach an empty directory name?
      EXIT LOOP                                               '// we are done, no more directories to process
    END IF
  LOOP                                                        '// walk through all of the directories found
  STDOUT "  Found: " + FORMAT$(fp, $longform) + " files in " + FORMAT$(dp, $longform) + " directories"
  REDIM PRESERVE f(fp)                                        '// shrink tables to just the size we need
  REDIM PRESERVE d(dp)
  STDOUT "Grouping files by size ..."                         '// tell user what we are doing just in case it takes a while
  ARRAY SORT f()                                              '// sort into file size order (the first parameter is size)
  i = 1
  DO WHILE i < fp - 1                                         '// cleaning up the table (must start at 1 more than minimum)
    DO WHILE f(i) = "" AND fp > 0                             '// is this a deleted file placeholder and we still have files in our list?
      debug ("File " + FORMAT$(i) + " is empty. FP = " + FORMAT$(fp))
      ARRAY DELETE f(i)                                       '// remove the empty element
      DECR fp                                                 '// lower the size of our table by one
    LOOP
    IF fp < 1 THEN                                            '// no more files to process?
      EXIT LOOP                                               '// leave the loop
    END IF
    buf = PARSE$(f(i),"|", 1)                                 '// get the file size
    DO WHILE PARSE$(f(i - 1),"|", 1) <> buf AND buf <> PARSE$(f(i + 1),"|", 1) '// this file is different size than previous and next?
      ARRAY DELETE f(i)                                       '// remove the different size file
      DECR fp                                                 '// lower the size of our table by one
      IF fp < 1 THEN                                          '// no more files to compare?
        EXIT LOOP                                             '// leave loop
      END IF
      buf = PARSE$(f(i),"|", 1)                               '// get the new file size
    LOOP
    INCR i                                                    '// move to the next file in the table
  LOOP
  IF f(0) = "" THEN                                           '// is first element empty?
    ARRAY DELETE f(0)                                         '// erase first element
    DECR fp                                                   '// move pointer down one
  END IF
  IF fp > 0 THEN                                              '// more than one file to compare?
    STDOUT "Scanning for duplicates in " + FORMAT$(fp + 1, $longform) + " file" + IIF$(fp + 1 <> 1, "s", "") + " ..."
    STDOUT "Finding CRC codes for files ..."
    FOR i = 0 TO UBOUND(f())
      IF LEN(f(i)) > 0 THEN
        f(i) = f(i) + "|" + crcfile(d(VAL(PARSE$(f(i), "|", 4))) + "\" + PARSE$(f(i), "|", 2)) '// add CRC information to end of file info
      END IF
    NEXT
    show dupe, fp - 1                                         '// change the title bar to show progress
    REDIM PRESERVE f(fp)                                      '// shrink the table to just the size we need
    FOR i = 0 TO fp                                           '// loop through the table
      j = INSTR(-1, f(i), "|")                                '// locate last pipe
      f(i) = MID$(f(i), j + 1) + "|" + LEFT$(f(i), j - 1)     '// put CRC first
    NEXT
    ARRAY SORT f()                                            '// sort into file CRC order
    FOR i = 0 TO fp                                           '// loop through the table
      j = INSTR(f(i)+"|","|")                                 '// locate first pipe
      f(i) = MID$(f(i),j + 1) + "|" + LEFT$(f(i),j - 1)       '// put CRC last
    NEXT
    IF verbose THEN
      STDOUT "[V] List of files to compare:"
      FOR i = 0 TO fp
        STDOUT "[V] " + FORMAT$(i,REPEAT$(LEN(FORMAT$(fp)),"0")) +"= "+ f(i)
      NEXT
      STDOUT ""
    END IF
    IF fp+1 > UBOUND(f()) THEN                                '// is pointer past the end of table?
      REDIM PRESERVE f(fp + 1)                                '// make sure we have enough room
    END IF
    FOR i = 0 TO fp                                           '// walk the list
      debug ("Working on record " + FORMAT$(i))
      IF f(i) = "" OR LEFT$(f(i), 1) = $dflag THEN            '// is this a deleted file placeholder?
        debug ("I think record " + FORMAT$(i) + " is deleted. Skipping to next file name.")
        ITERATE                                               '// try again with the next value for "i"
      END IF
      show dupe, fp - i                                       '// change the title bar to show progress
      k = 1                                                   '// start counting dupes with offset at first file
      j = i + 1                                               '// start with the next file
      debug ("I think the CRC match is " + IIF$(PARSE$(f(i), "|", 5) = PARSE$(f(j), "|", 5),"True","False"))
      DO WHILE j <= fp AND PARSE$(f(i), "|", 5) = PARSE$(f(j), "|", 5) '// loop through all of the CRC matches
        debug ("Testing record " + FORMAT$(i) + " with " + FORMAT$(j))
        IF LEFT$(f(j),1) <> $dflag THEN                       '// not deleted?
          fqueue = $CRLF + "The following files are binary duplicates." + $CRLF + " [1] " + unf(f(i)) + $CRLF '// remember original file name
          buf  = d(VAL(PARSE$(f(i), "|", 4))) + "\" + PARSE$(f(i), "|", 2) '// get filename 1 (path + file name)
          buf2 = d(VAL(PARSE$(f(j), "|", 4))) + "\" + PARSE$(f(j), "|", 2) '// get filename 2 (path + file name)
          IF same(buf, buf2) THEN                             '// are these the same file inside?
            debug ("Found duplicate: " + buf2)
            INCR dupe                                         '// count the number of dupe files we find
            INCR k
            fpump = fpump + " [" + FORMAT$(k) + "] " + unf(f(j)) + $CRLF '// add name to output queue
            IF flen(buf) > flen(buf2) THEN                    '// is file name 1 longer than file name 2?
              fpump = fpump + " (File " + FORMAT$(k) + tense + "erased because it had a shorter name.)" + $CRLF
              p = j                                           '// erase second file name
            ELSEIF flen(buf) < flen(buf2) THEN                '// is file name 1 longer than file name 2?
              fpump = fpump + " (File " + FORMAT$(k) + tense + "erased because it had a shorter name.)" + $CRLF
              p = i                                           '// erase first file
            ELSEIF TALLY(buf, "\") < TALLY(buf2, "\") THEN    '// is path 1 shorter than path 2?
              fpump = fpump + " (File " + FORMAT$(k) + tense + "erased because it had a shorter path.)" + $CRLF
              p = i
            ELSEIF TALLY(buf, "\") > TALLY(buf2, "\") THEN    '// is path 2 shorter than path 1?
              fpump = fpump + " (File " + FORMAT$(k) + tense + "erased because it had a shorter path.)" + $CRLF
              p = j
            ELSE                                              '// the paths are the same depth
              fpump = fpump + " (File " + FORMAT$(k) + tense + "erased.)" + $CRLF '// we just have to pick a name to erase, why not this one
              p = j
            END IF                                            '// which file name is longer
            freespace = freespace + VAL(PARSE$(f(p), "|", 1)) '// recover this many bytes
            buf2 = d(VAL(PARSE$(f(p), "|", 4))) + "\" + PARSE$(f(p), "|", 2)
            fKILL buf2                                        '// erase the selected file
            MID$(f(p), 1) = $dflag                            '// mark record deleted
            freespace = freespace + VAL(PARSE$(f(j), "|", 1)) '// recover this many bytes
          END IF                                              '// are these the same file inside
        ELSE                                                  '// this is marked deleted
          debug (d(VAL(PARSE$(f(j), "|", 4))) + "\" + PARSE$(f(j), "|", 2) + " is marked deleted.")
        END IF
        INCR j                                                '// move to the next file
      LOOP
      IF LEN(fpump) > 0 THEN                                  '// do we have anything to log?
        flog fqueue + fpump                                   '// write our entry to the log file
        fpump = ""
      END IF
    NEXT
    buf = "Found " + IIF$(dupekill, "and removed ", "") + FORMAT$(dupe, $longform) + " duplicate" + IIF$(dupe = 1,"","s") + "."
  ELSE
    buf = "No files found with the same file size."
  END IF                                                      '// something to find
  STDOUT buf                                                  '// tell user what we found
  IF dupe > 0 AND exist(logfile) THEN                         '// is the log file here?
    flog $CRLF + RIGHT$(STRING$(70,"-") + "[" + DATE$ + " " + TIME$ + "]---", 70) + $CRLF + buf '// record it to the log file
  END IF
  IF freespace > 0 THEN
    buf = FORMAT$(freespace, $longform) + " byte" + IIF$(freespace <> 1, "s","") + " of free disk space " + IIF$(dupekill,"were","would have been") + " recovered."
    STDOUT buf                                                '// tell user about space recovered
    IF dupe > 0 AND exist(logfile) THEN                       '// is the log file here?
      flog $CRLF + buf + $CRLF                                '// record it to the log file too
    END IF
  END IF
  IF dupe > 0 AND exist(logfile) THEN                         '// is the log file here?
    STDOUT "The duplicate files are listed in " + $DQ + logfile + $DQ
  END IF
  IF verbose THEN
    i = ABS(INT(TIMER - tstart))
    debug ("Run time: " + FORMAT$(i) + " second" + IIF$(i<>1,"s","") + ".")
  END IF
END FUNCTION
